/*
    Solar - a solar system simulator
    Copyright (C) 2004, 2005  Sam Morris <sam@robots.org.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License, version 2, as
    published by the Free Software Foundation.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "Planet.h"

#include <cassert>
#include <cmath>
#include <iostream>

using namespace solar;

Planet::Planet(std::string name) : radius(0), mass(0), position(0, 0, 0), velocity(0, 0, 0), colour(0.5, 0.5, 0.5)
{
        this->name = name;
}

Planet::Planet(xmlpp::Node *node): position(0, 0, 0), velocity(0, 0, 0), colour(0.5, 0.5, 0.5)
{
        xmlpp::Element *e = dynamic_cast<xmlpp::Element *>(node);

        this->name   =       e->get_attribute("name")->get_value()    ;
        this->radius = atof( e->get_attribute("radius")->get_value().data() );
        this->mass   = atof( e->get_attribute("mass")->get_value().data()   );

        this->position = Vector( dynamic_cast<xmlpp::Element *>(node->find("position")[0]) );
        this->velocity = Vector( dynamic_cast<xmlpp::Element *>(node->find("velocity")[0]) );
        this->colour   = Vector( dynamic_cast<xmlpp::Element *>(node->find("colour")[0]   ) );
}

const Vector& Planet::getPosition() const
{
        return this->position;
}

void Planet::setPosition(const Vector& position)
{
        this->position = position;
}


const Vector& Planet::getVelocity() const
{
        return this->velocity;
}

void Planet::setVelocity(const Vector& velocity)
{
        this->velocity = velocity;
}


const Vector& Planet::getColour() const
{
        return this->colour;
}

void Planet::setColour(const Vector& colour)
{
        this->colour = colour;
}


void Planet::setMass(float mass)
{
        this->mass = mass;
}

float Planet::getMass() const
{
        return this->mass;
}

void Planet::setRadius(float radius)
{
        this->radius = radius;
}

float Planet::getRadius() const
{
        return this->radius;
}

const std::string& Planet::getName() const
{
        return this->name;
}

void Planet::setName(const std::string& name)
{
        this->name = name;
}

/**
 * Returns the force of gravitational attraction between two bodies in Newtons.
 */
Vector Planet::getGravity(const Planet& planet) const
{
        float m1 = this->getMass();
        float m2 = planet.getMass();

        /**
         * r is the vector that takes is from 'planet' to *this.
         */
        Vector r = planet.getPosition() - this->getPosition();

        if (m1 == 0 || m2 == 0 || r.getMagnitude() == 0)
                return Vector(0, 0, 0);

        double force = (solar::G * m1 * m2) / (pow(r.getMagnitude(), 2));

        vector_t Fx = force * cos( r.getTheta() );
        vector_t Fy = force * cos( r.getPhi() );
        vector_t Fz = force * cos( r.getPsi() );
        
        assert(Fx != INFINITY);
        assert(Fy != INFINITY);
        assert(Fz != INFINITY);
        return Vector(Fx, Fy, Fz);
}

std::ostream& operator <<(std::ostream& lhs, const Planet& rhs)
{
        lhs << rhs.getName() << " {mass=" << rhs.getMass() << ", radius=" << rhs.getRadius() << ", position=" << rhs.getPosition() << ", velocity=" << rhs.getVelocity() << ", colour=" << rhs.getColour() << "}";
        return lhs;
}
