/*
    Solar - a solar system simulator
    Copyright (C) 2004, 2005  Sam Morris <sam@robots.org.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License, version 2, as
    published by the Free Software Foundation.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "Scene.h"

#include <cmath>

#include "Planet.h"
#include "PlanetWindow.h"

using namespace solar;

Scene::Scene(Universe& universe)
        : exaggeration(EXAGGERATE_LOG),
          zoomDistance(2.5e11),
          frameCount(0)
{
        this->universe = &universe;

        if (universe.planets.empty() == true)
                followIterator_valid = false;
        else
        {
                followIterator = universe.planets.begin(); //TODO: handle empty universe
                followIterator_valid = true;
        }

        // Configure quadric "paintbrush" for planets
        planetQuad = gluNewQuadric();
        gluQuadricDrawStyle(   planetQuad, GLU_FILL );
        gluQuadricNormals(     planetQuad, GLU_SMOOTH );
        gluQuadricOrientation( planetQuad, GLU_OUTSIDE );
        
        // Configure double-buffered OpenGL visual with z-buffer
        Glib::RefPtr<Gdk::GL::Config> glconfig;
        glconfig = Gdk::GL::Config::create(Gdk::GL::MODE_RGB   |
                                           Gdk::GL::MODE_DEPTH |
                                           Gdk::GL::MODE_DOUBLE);
        // if (glconfig.is_null())
        if (! glconfig)
        {
                std::cerr << "Couldn't initialise OpenGL visual." << std::endl;
                std::exit(1); //TODO: fail initialisation?
        }

        set_gl_capability(glconfig);
        
        // Tell GTK that we want to recieve notifications for these events
        this->add_events(Gdk::VISIBILITY_NOTIFY_MASK);
        this->add_events(Gdk::BUTTON_RELEASE_MASK);
        this->add_events(Gdk::BUTTON_PRESS_MASK); // we don't recieve RELEASE if we don't recieve PRESS
}

Scene::~Scene()
{
        gluDeleteQuadric(planetQuad);
}

/**
 * Called "when the widget's GDK resources are created" as best as I can discern.
 * Gtk::Widget documentation is sparse :)
 */
void Scene::on_realize()
{
        // Call base on_realize()
        Gtk::DrawingArea::on_realize();

        Glib::RefPtr<Gdk::GL::Drawable> gldrawable = get_gl_drawable();
        if (!gldrawable->gl_begin(get_gl_context()))
        {
                std::cerr << "Couldn't begin OpenGL." << std::endl;
                return;
        }

        /*
        std::cerr << "GL_RENDERER   = " << glGetString(GL_RENDERER)   << std::endl;
        std::cerr << "GL_VERSION    = " << glGetString(GL_VERSION)    << std::endl;
        std::cerr << "GL_VENDOR     = " << glGetString(GL_VENDOR)     << std::endl;
        //std::cerr << "GL_EXTENSIONS = " << glGetString(GL_EXTENSIONS) << std::endl;
        std::cerr << std::endl;
        */

        gldrawable->gl_end();

        timer.start();
}

/**
 * Called on resize
 */
bool Scene::on_configure_event(GdkEventConfigure* event)
{
        // Configure the Projection matrix with the size of window etc

        Glib::RefPtr<Gdk::GL::Drawable> gldrawable = get_gl_drawable();
        if (!gldrawable->gl_begin(get_gl_context()))
                return false;


        /* 
        glMatrixMode(GL_PROJECTION);
        glLoadIdentity();

        //GLfloat h = (GLfloat)(get_height()) / (GLfloat)(get_width());
        //glFrustum(-1.0, 1.0, -h, h, 0.1, 60.0);
        //
        glOrtho(-zoomDistance, zoomDistance, -zoomDistance, zoomDistance, -zoomDistance, zoomDistance);
        */ 
        glViewport(0, 0, get_width(), get_height());
        
        
        gldrawable->gl_end();
        return true;
}

/**
 * Called when the widget needs to be drawn because eg, another window
 * moved away
 */
bool Scene::on_expose_event(GdkEventExpose* event)
{
        float axisLength = zoomDistance/8;
        float boundLength = zoomDistance/64;

        // Draw scene
        Glib::RefPtr<Gdk::GL::Drawable> gldrawable = get_gl_drawable();
        if (!gldrawable->gl_begin(get_gl_context()))
                return false;

        glMatrixMode(GL_PROJECTION);
        glLoadIdentity();
        glOrtho(-zoomDistance, zoomDistance, -zoomDistance, zoomDistance, -zoomDistance, zoomDistance);
        
        // Draw planets
        glMatrixMode(GL_MODELVIEW);
        glLoadIdentity();

        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

        // Move "camera" to centre on the followed planet
        glTranslatef(-getCameraPosition().getX(), -getCameraPosition().getY(), -getCameraPosition().getZ());

        std::vector<Planet>::iterator i;
        for (i = universe->planets.begin(); i != universe->planets.end(); i++)
        {       
                glPushMatrix();

                Vector position = i->getPosition();               
                float radius = exaggerateRadius(i->getRadius());

                glTranslatef(position.getX(), position.getY(), position.getZ());

                glColor3f(i->getColour().getX(), i->getColour().getY(), i->getColour().getZ());
               
                // draw planet box
                glBegin(GL_LINE_LOOP);
                glVertex3f(-boundLength, -boundLength, 0);
                glVertex3f(+boundLength, -boundLength, 0);
                glVertex3f(+boundLength, +boundLength, 0);
                glVertex3f(-boundLength, +boundLength, 0);
                glEnd();

                // draw planet body
                gluSphere(planetQuad, radius, 15, 15);

                // draw velocity line
                if (i->getVelocity().getMagnitude() > 1) // don't draw if we are almost stationary
                {
                        Vector vLine = i->getVelocity().toUnitVector() * axisLength;

                        glBegin(GL_LINES);
                        glVertex3f(0, 0, 0);
                        glVertex3f(vLine.getX(), vLine.getY(), vLine.getZ());
                        glEnd();
                }
                

                glPopMatrix();
        }
       

        //glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
        glBegin(GL_LINES);

        glColor3f(1, 0, 0);
        glVertex3f(0, 0, 0);
        glVertex3f(axisLength, 0, 0);

        glColor3f(0, 1, 0);
        glVertex3f(0, 0, 0);
        glVertex3f(0, axisLength, 0);
        glEnd();
        
        glColor3f(0, 0, 1);
        glVertex3f(0, 0, 0);
        glVertex3f(0, 0, axisLength);

        glEnd();


        // Swap buffers.
        gldrawable->swap_buffers();

        gldrawable->gl_end();


        // Print frame rate.
        frameCount++;

        double seconds = timer.elapsed();
        if (seconds >= 1.0)
        {
                /*
                std::cout.setf(std::ios::fixed, std::ios::floatfield);
                std::cout.precision(0);
                std::cout << "Scene: " << frameCount << " frames / " << seconds << " secs = " << (frameCount / seconds) << " FPS\n";
*/
                timer.reset();
                frameCount = 0;
        }

        return true;
}

/**
 * Called when a window becomes hidden, obscured etc
 *
 * If the window is no longer visible, we disconnect the idle callback
 * so that we don't waste CPU time drawing the scene
 */
bool Scene::on_visibility_notify_event(GdkEventVisibility* event)
{
        if (event->state == GDK_VISIBILITY_FULLY_OBSCURED)
        {
                if (idleConnection.connected())
                        idleConnection.disconnect();
        }
        else
        {
                if (!idleConnection.connected())
                        idleConnection = Glib::signal_idle().connect( sigc::mem_fun(*this, &Scene::on_idle), GDK_PRIORITY_REDRAW );
        }

        return true;
}

/**
 * Called by the main GTK loop when there are no other pending events.
 * We simple trigger a redraw of the scene.
 */
bool Scene::on_idle()
{
        this->queue_draw();

        return true;
}

bool Scene::on_button_release_event(GdkEventButton* event)
{
        if (universe->getRunning() == false)
        {
                double x = (event->x / this->get_width());
                x = (x - 0.5) * 2;
                x *=  zoomDistance;
                
                double y = (event->y / this->get_height());
                y = (y - 0.5) * 2;
                y *= -zoomDistance; // -ve cos screen y axis is upside down

                Vector clickPosition = Vector(x, y, zoomDistance) + getCameraPosition();
                //std::cout << "click at " << clickPosition << std::endl;

                for (std::vector<Planet>::iterator i = universe->planets.begin(); i != universe->planets.end(); i++)
                {
                        if ( universe->collideRaySphere( clickPosition, Vector(0, 0, -1), i->getPosition(), exaggerateRadius(i->getRadius())) )
                        {
                                PlanetWindow w( &(*i) );

                                if (w.run())
                                        w.updatePlanet( *i );
                        }
                }
        }

        return true;
}

bool Scene::on_button_press_event(GdkEventButton* event)
{
        return true;
}

exaggerateMode Scene::getExaggerateMode() const
{
        return this->exaggeration;
}

void Scene::setExaggerateMode(exaggerateMode exaggeration)
{
        this->exaggeration = exaggeration;
}

double Scene::exaggerateRadius(double radius) const
{
        switch (getExaggerateMode())
        {
        case EXAGGERATE_LOG:
                radius = log(radius) * 0.5e9;
                break;
        
        case EXAGGERATE_NONE:
                break;

        }

        return radius;
}

double Scene::getZoomDistance() const
{
        return this->zoomDistance;
}

void Scene::setZoomDistance(double zoomDistance)
{
        this->zoomDistance = zoomDistance;
}

void Scene::followNextPlanet()
{
        if (followIterator < universe->planets.end()-1)
                followIterator++;
        else
                followIterator = universe->planets.begin();
}

void Scene::followPrevPlanet()
{
        if (followIterator > universe->planets.begin())
                followIterator--;
        else
                followIterator = universe->planets.end()-1;
}

Vector Scene::getCameraPosition()
{
        if (universe->planets.empty() == false)
        {
                if (followIterator_valid == true)
                        return followIterator->getPosition();
                else
                {
                        followIterator_valid = true;
                        followIterator = universe->planets.begin(); //TODO: handle empty universe
                        return getCameraPosition();
                }
        }
        else
                return Vector(0, 0, 0);
}
