/*
    Solar - a solar system simulator
    Copyright (C) 2004, 2005  Sam Morris <sam@robots.org.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License, version 2, as
    published by the Free Software Foundation.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "Universe.h"

#include <cassert>
#include <iostream>
#include <vector>
#include <gtkmm.h>
#include <cmath>

#include "toString.h"

using namespace solar;

Universe::Universe(const std::string& filename, double timeStep): running(false), timeMult(1), timeStep(0), lastT(0.0), elapsedTime(0), timeIntegrationThreshold(60 * 60 * 12)
{
        this->timeStep = timeStep;
        this->timer = new Glib::Timer();

        if (filename != "")
        {
                xmlpp::DomParser parser(filename);
				assert (parser);
                const xmlpp::Node* root = parser.get_document()->get_root_node();
				assert (root != 0);

                xmlpp::NodeSet set = root->find("/solar/planets/planet");

                for (xmlpp::NodeSet::iterator i = set.begin(); i != set.end(); i++)
                {
                        Planet p(*i);
                        this->add(p);
                }
        }
}

Universe::~Universe()
{
        // LEAK: if deleted, all subsequent timers cause segfaults
        //delete this->timer;
        this->timer = NULL;
}

/**
 * Adds planets to represent Sol's inner system; used for testing until we
 * get loading from a file working.
 */
/*
void Universe::addDefaultPlanets()
{
        Planet sun = Planet("sol");
        sun.setColour( Vector(1, 1, 0) );
        sun.setRadius(6.95e8);
        sun.setMass(1.989e30);
        add(sun);
        
        Planet earth = Planet("earth");
        earth.setColour( Vector(0, 0, 1) );
        earth.setRadius(6.38e6);
        earth.setMass(5.98e24);
        earth.setPosition( Vector(1.49e11 , 0, 0) ); // km
        earth.setVelocity(Vector(0,
        /+        0.1 * sqrt(
                        (6.67e-11 * earth.getMass())
                                /
                        earth.getPosition().getMagnitude()
                )+/
                29.973e3, 0));
        add(earth);

        Planet moon = Planet("luna");
        moon.setColour( Vector(0.8, 0.8, 0.8) );
        moon.setRadius(1737);
        moon.setMass(7.34e22);
        moon.setPosition( earth.getPosition() - Vector(3.84e8, 0, 0) );
        moon.setVelocity( earth.getVelocity() + Vector(0, 1028, 0) );
        add(moon);

        Planet mercury = Planet("mercury");
        mercury.setColour( Vector(0.7, 0.3, 0.7) );
        mercury.setRadius(2.57e6);
        mercury.setMass(3.28e23);
        mercury.setPosition( Vector(5.79e10, 0, 0) );
        mercury.setVelocity(Vector(0, 49.87e3, 0) );
        add(mercury);

        Planet venus = Planet("venus");
        venus.setColour( Vector(1.0, 1.0, 0.5) );
        venus.setRadius(6.31e6);
        venus.setMass(4.38e24);
        venus.setPosition( Vector(1.08e11, 0, 0) );
        venus.setVelocity( Vector(0, 35.02e3, 0));
        add(venus);

        Planet mars = Planet("mars");
        mars.setColour( Vector(1.0, 0.5, 0.5) );
        mars.setRadius(3.43e6);
        mars.setMass(6.37e23);
        mars.setPosition( Vector(2.28e11, 0, 0) );
        mars.setVelocity(Vector(0, 24.13e3, 0));
        add(mars);
}
*/

void Universe::add(Planet& p)
{
        planets.push_back(p);
}

long double Universe::getElapsedTime() const
{
        return elapsedTime;
}

/**
 * First, we take a reading from the timer, and store it in seconds.
 * deltaT records the time difference between this frame and the last.
 */
bool Universe::think()
{
        long double seconds;
        long double deltaT;

        if (timeStep == 0)
        {       // realtime mode
                seconds = timer->elapsed();      // read timer
                deltaT = (seconds - lastT) * getTimeMult();
                lastT = seconds;
        }
        else
                deltaT = timeStep;

//std::cout << "seconds=" << seconds << "\tlastT=" << lastT << "\tdeltaT=" << deltaT << std::endl;

        elapsedTime += deltaT;
        //assert (elapsedTime >= 0);

        if (timeStep == 0)
        {       // realtime mode
                double partialDeltaT = timeIntegrationThreshold;
                while (deltaT != 0) //TODO: handle negative T; slow down sim on slow computes
                {
                        //double loopStartTime = timer->elapsed();

                        if (deltaT > 0 && deltaT >= timeIntegrationThreshold)
                        {
                                deltaT -= partialDeltaT;
                                //std::cout << "\t splitting (" << deltaT << " s left)" << std::endl;
                        }
                        else if (deltaT < 0 && deltaT <= -timeIntegrationThreshold)
                        {
                                deltaT -= -partialDeltaT;
                                //std::cout << "\t splitting (" << deltaT << " s left)" << std::endl;
                        }
                        else
                        {
                                //std::cout << "\t FINAL RUN (" << deltaT << " s)" << std::endl << std::endl; 
                                partialDeltaT = deltaT;
                                deltaT = 0;
                        }

                        doGravity(partialDeltaT);
                }
        }
        else
                doGravity(deltaT);
        
        // Calculate FPS; TODO: make this not fsck up elapsedTime
        /*frameCount++;
        if (seconds > 1.0)
        {
                //std::cout.setf(std::ios::fixed, std::ios::floatfield);
                //std::cout.precision(0);
                //std::cout << "Universe: " << frameCount << " frames / " << seconds << " secs = ";
                std::cout << (frameCount/seconds) << " FPS" << std::endl;

                timer->reset();
                frameCount = 0;
        }*/

        return true;
}

void Universe::doGravity(double deltaT)
{
                /**
                 * For every planet x
                 *      f = [0,0,0]
                 *      for every other planet y
                 *              f += gravity(x, y)
                 *      acceleration = f/m
                 *      velocity += acceleration * deltaT
                 *      position += velocity * deltaT
                 */
                std::vector<Planet> planets_new; // holds new universe state

        //std::cout << "THINK" << std::endl << std::endl;
                
                for (std::vector<Planet>::iterator i = planets.begin(); i != planets.end(); i++)
                {
        //std::cout << (*i) << std::endl;
                        Vector force(0.0, 0.0, 0.0);

                        for (std::vector<Planet>::iterator j = planets.begin(); j != planets.end(); j++)
                        {
                                if (i == j) continue;

                                force = force + i->getGravity( (*j) );
                        }
        //std::cout << "\tforce = " << force << std::endl;

                        // a = f/m
                        Vector acceleration(0, 0, 0);
                        if (i->getMass() != 0)
                                acceleration = force * ( 1/ i->getMass() );

        //std::cout << "\tacceleration = " << acceleration << std::endl;
                        Planet temp = (*i); // copy planet

        //std::cout << "\tdeltaT = " << deltaT << std::endl;

                        // v += at
                        Vector newVelocity = temp.getVelocity() + acceleration * deltaT;
        //std::cout << "\tnewVelocity = " << newVelocity << std::endl;
                        temp.setVelocity( newVelocity );

                        // s += ut
                        Vector newPosition = temp.getPosition() + temp.getVelocity() * deltaT;
        //std::cout << "\tnewPosition = " << newPosition << std::endl;
                        temp.setPosition( newPosition );
                        
                        planets_new.push_back( temp );
        //std::cout << std::endl;
                }

                // collision detection over planets_new

                // copy new universe state over old
                planets = std::vector<Planet>(planets_new);

                // reset position of voidlanets to centre
                if (planets.empty() == false)
                {
                        Vector sunPosition = planets[0].getPosition();
                        for (std::vector<Planet>::iterator i = planets.begin(); i != planets.end(); i++)
                                i->setPosition(i->getPosition() - sunPosition);
                }

                //if (timer->elapsed()-loopStartTime > 10)
                //{
                //        std::cerr << "Your computer is too slow!" << std::endl;
                //}
}

bool Universe::getRunning() const
{
        return running;
}

void Universe::setRunning(bool running)
{
        if (running == this->running) //avoid multiple connections to signal
                return;

        if (running)
        {
                thinkConnection = Glib::signal_idle().connect( sigc::mem_fun(*this, &Universe::think), GDK_PRIORITY_REDRAW );
                timer->reset();
                timer->start();
        }
        else
        {
                thinkConnection.disconnect();
                timer->stop();
                lastT = 0;
        }

        this->running = running;
}

void Universe::printStatus()
{
        std::cout << "Universe contents:" << std::endl;
        std::cout << std::scientific;
        std::cout.precision(2);

        for (std::vector<Planet>::iterator i = planets.begin(); i != planets.end(); i++)
                std::cout << '\t' << *i << " @" << &(*i) << std::endl;

        std::cout << std::endl;
}

void Universe::saveToFile(const std::string& filename)
{
        xmlpp::Document document;
        xmlpp::Node *rootNode = document.create_root_node("solar");

        xmlpp::Element *planetsNode = rootNode->add_child("planets");

        for (std::vector<Planet>::iterator i = planets.begin(); i != planets.end(); i++)
        {
                xmlpp::Element *planetNode = planetsNode->add_child("planet");

                // name
                planetNode->set_attribute("name", i->getName());

                // colour
                xmlpp::Element *colourNode = planetNode->add_child("colour");
                i->getColour().addAttrsToXmlElement(colourNode);
                
                // radius
                planetNode->set_attribute("radius", toString( i->getRadius() ));

                // mass
                planetNode->set_attribute("mass", toString( i->getMass() ));

                // position
                xmlpp::Element *positionNode =planetNode->add_child("position");
                i->getPosition().addAttrsToXmlElement(positionNode);
                
                // velocity
                xmlpp::Element *velocityNode =planetNode->add_child("velocity");
                i->getVelocity().addAttrsToXmlElement(velocityNode);
        }

        document.write_to_file(filename);
}

double Universe::getTimeMult() const
{
        return this->timeMult;
}

void Universe::setTimeMult(double timeMult)
{
        this->timeMult = timeMult;
}

bool Universe::collideRaySphere(const Vector& rStart, const Vector& rDirection, const Vector& sPosition, double sRadius)
{
        Vector Q = sPosition - rStart;
        vector_t length = Q.getMagnitude();
        double v = Q.dotProduct( rDirection.toUnitVector() );
        double d = sRadius*sRadius - ( length*length - v*v);

        return (d >= 0);
}
