/*
    Solar - a solar system simulator
    Copyright (C) 2004, 2005  Sam Morris <sam@robots.org.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License, version 2, as
    published by the Free Software Foundation.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef SOLAR_UNIVERSE_H
#define SOLAR_UNIVERSE_H

/**
 * A class to model the universe.
 *
 * Inherits from SigC::Object so that the think() method can act as a signal
 * handler.
 */

#include "Planet.h"

#include <vector>
#include <libxml++/libxml++.h>
#include <gtkmm.h>

namespace solar {

/**
  * Gravitational Constant G = 6.673 × N m^2 kg^-1
  * Two 1 kg masses, 1 m apart, have a force of ~6.67 × 10−11 N
  * between them.
  */
const double G = 6.673e-11;

class Universe : public sigc::trackable
{
public:
        Universe(const std::string& filename, double timeStep = 0);
        ~Universe();

        void printStatus(); // not const 'cos of iterator
        void saveToFile(const std::string& filename); //ditto

        //void addDefaultPlanets();
        void add(Planet& p);

        bool getRunning() const;
        void setRunning(bool running);
        bool think();

        double getTimeMult() const;
        void setTimeMult(double timeMult);

        long double getElapsedTime() const;

        /**
         * Calculates whether a ray intersects with a sphere.
         * @return true if the ray intersects the sphere.
         */
        bool collideRaySphere(const Vector& rStart, const Vector& rDirection, const Vector& sPosition, double sRadius);

        std::vector<Planet> planets; // should really be private
private:
        
        bool running;
        SigC::Connection thinkConnection;

        void doGravity(double deltaT);

        double timeMult; // factor to compress time by
        double timeStep; // used for batch simulation
        Glib::Timer *timer;
        long double lastT; // time of last think(); used to discover deltaT
        //int frameCount;
        long double elapsedTime;

        int timeIntegrationThreshold;
};

}
#endif
