/*
    Solar - a solar system simulator
    Copyright (C) 2004, 2005  Sam Morris <sam@robots.org.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License, version 2, as
    published by the Free Software Foundation.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "Vector.h"

#include <cmath>
#include <cassert>

#include "toString.h"

using namespace solar;

Vector::Vector(vector_t x, vector_t y, vector_t z)
{
        this->setX(x);
        this->setY(y);
        this->setZ(z);
}

Vector::Vector(xmlpp::Element *e)
{
        this->setX(atof( e->get_attribute("x")->get_value().data()));
        this->setY(atof( e->get_attribute("y")->get_value().data()));
        this->setZ(atof( e->get_attribute("z")->get_value().data()));
}


vector_t Vector::getX() const
{
        return this->x;
}

vector_t Vector::getY() const
{
        return this->y;
}

vector_t Vector::getZ() const
{
        return this->z;
}


void Vector::setX(vector_t x)
{
        this->x = x;
}

void Vector::setY(vector_t y)
{
        this->y = y;
}

void Vector::setZ(vector_t z)
{
        this->z = z;
}


vector_t Vector::getMagnitude() const
{
        return sqrt(
                  pow(getX(), 2)
                + pow(getY(), 2)
                + pow(getZ(), 2)
        );       //TODO: use accessors?
}

/**
 * The dot product of two *unit* vectors is the cosine of the angle between
 * them.
 */
double Vector::dotProduct(const Vector& vector) const
{
        Vector v = *this;
        Vector w = vector;

        return    v.getX() * w.getX()
                + v.getY() * w.getY()
                + v.getZ() * w.getZ();
        
        /*
        //assert(result <=  1);
        //      assert(result >= -1);
        if (result > 1)
        {
                //std::cerr << "Vector::doProduct: clamped " << result << " to 1." << std::endl;
                return 1;
        }
        else if (result < -1)
        {
               // std::cerr << "Vector::doProduct: clamped " << result << " to -1." << std::endl;
                return -1;
        }
        else
                return result;
        */
}

/**
 * The cross product of two vectors is a vector orthagonal to the other two.
 * This operator is not commutative: for " v×w ", use " v.crossProduct(w) "
 * Postcondition: |v•w| = |v| |w| sin(theta)
 */
Vector Vector::crossProduct(const Vector& vector) const
{
        Vector v = *this;
        Vector w = vector;
        
        vector_t x = v.getY() * w.getZ() - v.getZ() * w.getY();
        vector_t y = v.getZ() * w.getX() - v.getX() * w.getZ();
        vector_t z = v.getX() * w.getY() - v.getY() * w.getX();

        return Vector(x, y, z);
}

Vector Vector::toUnitVector() const
{
        Vector result = Vector(
                this->getX() / this->getMagnitude(),
                this->getY() / this->getMagnitude(),
                this->getZ() / this->getMagnitude()
        );

        //assert(result.getMagnitude() == 1);
        return result;
}

// Operators
// 

Vector operator -(const Vector& vector)
{
        return Vector(
                0 - vector.getX(),
                0 - vector.getY(),
                0 - vector.getZ()
        );
}

Vector operator +(const Vector& lhs, const Vector& rhs)
{       
        vector_t x = lhs.getX() + rhs.getX();
        vector_t y = lhs.getY() + rhs.getY();
        vector_t z = lhs.getZ() + rhs.getZ();

        return Vector(x, y, z);
}

Vector operator -(const Vector& lhs, const Vector& rhs)
{
        Vector foo = -rhs;
        return lhs + foo;
}

Vector operator *(const Vector& lhs, vector_t rhs)
{
        return Vector(lhs.getX() * rhs, lhs.getY() * rhs, lhs.getZ() * rhs);
}

std::ostream& operator <<(std::ostream& lhs, const solar::Vector& rhs)
{
        lhs << "{" << rhs.getX() << ", " << rhs.getY() << ", " << rhs.getZ() << "}";
        return lhs;
}

/**
 * Returns the angle, in radians, between the vector and the X axis.
 */
double Vector::getTheta() const
{
        Vector i = Vector(1, 0, 0);
        double theta = acos( this->toUnitVector().dotProduct(i) );
        
        if (this->getY() >= 0)
                return theta;
        else // fix to make function return 0 <= theta < 2 * M_PI
                return M_PI + M_PI-theta;
}

/**
 * Returns the angle, in radians, between the vector and the Y axis.
 */
double Vector::getPhi() const
{
        Vector j = Vector(0, 1, 0);
        double phi = acos( this->toUnitVector().dotProduct(j) ); 

        if (this->getX() >= 0)
                return phi;
        else
                return M_PI + M_PI-phi;
}

/**
 * Returns the angle, in radians, between the vector and the Z axis.
 */
double Vector::getPsi() const
{
        Vector k = Vector(0, 0, 1);
        double psi = acos( this->toUnitVector().dotProduct(k) ); 

        if (this->getX() >= 0)
                return psi;
        else
                return M_PI + M_PI-psi;
}

/**
 * Converts the vector into a string of the form:
 *      <$name x='$x' y='$y' z='$z'/>
 */
/*std::string toXmlNodeString(const std::string& name) const
{
        std::stringstream os;
        os << '<' << name << " x='" << getX() << "' y='" << getY() << "' z='" << getZ() << "'/>";
        return os.str();
}*/

/**
 * Adds x, y, z attributes to the given Xml Element node.
 */
void Vector::addAttrsToXmlElement(xmlpp::Element *colourNode) const
{
        colourNode->set_attribute("x", toString(this->getX()));
        colourNode->set_attribute("y", toString(this->getY()));
        colourNode->set_attribute("z", toString(this->getZ()));
}
